@@ -4,6 +4,7 @@
 ##
 ## PySol -- a Python Solitaire game
 ##
+## Copyright (C) 2000 Markus Franz Xaver Johannes Oberhumer
 ## Copyright (C) 1999 Markus Franz Xaver Johannes Oberhumer
 ## Copyright (C) 1998 Markus Franz Xaver Johannes Oberhumer
 ##
@@ -23,28 +24,38 @@
 ## 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 ##
 ## Markus F.X.J. Oberhumer
-## <markus.oberhumer@jk.uni-linz.ac.at>
-## http://wildsau.idv.uni-linz.ac.at/mfx/pysol.html
+## <markus@oberhumer.com>
+## http://www.oberhumer.com/pysol
 ##
 ##---------------------------------------------------------------------------##
 
 
 # imports
-import os, re, sys, string, time
+import os, re, sys, string, time, types
 
 # PySol imports
-from mfxutil import EnvError, SubclassResponsibility                #bundle#
-from mfxutil import Struct, destruct                                #bundle#
-from random import constructRandom                                  #bundle#
-from stats import PysolStatsWriter                                  #bundle#
+from mfxtools import *
+from mfxutil import EnvError, SubclassResponsibility
+from mfxutil import Struct, destruct, openURL
+from util import PACKAGE, PACKAGE_URL, VERSION, bundle
+from random import constructRandom
+
+# stats imports
+from stats import Status_StatsDialog, PysolStatsFormatter
+from pysoltk import SingleGame_StatsDialog, AllGames_StatsDialog
+from pysoltk import FullLog_StatsDialog, SessionLog_StatsDialog
 
 # toolkit imports
-from pysoltk import EVENT_HANDLED, EVENT_PROPAGATE                  #bundle#
-from pysoltk import MfxDialog, MfxSimpleSlider, MfxSimpleEntry      #bundle#
-from pysoltk import MfxExceptionDialog                              #bundle#
-from pysoltk import MfxCheckMenuItem, MfxRadioMenuItem              #bundle#
-from pysoltk import getFont                                         #bundle#
-from help import helpAbout, helpHTML                                #bundle#
+from pysoltk import EVENT_HANDLED, EVENT_PROPAGATE
+from pysoltk import MfxDialog, MfxSimpleSlider, MfxSimpleEntry
+from pysoltk import MfxExceptionDialog
+from pysoltk import MfxCheckMenuItem, MfxRadioMenuItem
+from pysoltk import getFont
+from pysoltk import PlayerOptionsDialog
+from pysoltk import SoundOptionsDialog
+from pysoltk import DemoOptionsDialog, HintOptionsDialog
+from pysoltk import EditTextDialog
+from help import helpAbout, helpAboutSimple, helpHTML
 
 
 # /***********************************************************************
@@ -56,10 +67,11 @@ class PysolMenubarActions:
         self.app = app
         self.top = top
         self.game = None
-        # this is set by updateMenuState()
+        # enabled/disabled - this is set by updateMenuState()
         self.menustate = Struct(
             save = 0,
             save_as = 0,
+            hold_and_quit = 0,
             undo = 0,
             redo = 0,
             restart = 0,
@@ -77,24 +89,25 @@ class PysolMenubarActions:
         self.tkopt = Struct(
             gameid = MfxRadioMenuItem(self),
             gameid_popular = MfxRadioMenuItem(self),
-            confirm = MfxCheckMenuItem(self),
+            comment = MfxCheckMenuItem(self),
             autofaceup = MfxCheckMenuItem(self),
             autodrop = MfxCheckMenuItem(self),
             autodeal = MfxCheckMenuItem(self),
             quickplay = MfxCheckMenuItem(self),
             undo = MfxCheckMenuItem(self),
+            bookmarks = MfxCheckMenuItem(self),
             hint = MfxCheckMenuItem(self),
             highlight_piles = MfxCheckMenuItem(self),
             highlight_cards = MfxCheckMenuItem(self),
             highlight_samerank = MfxCheckMenuItem(self),
             sound = MfxCheckMenuItem(self),
-            cardset = MfxRadioMenuItem(self),
             cardback = MfxRadioMenuItem(self),
             tabletile = MfxRadioMenuItem(self),
             animations = MfxRadioMenuItem(self),
             shadow = MfxCheckMenuItem(self),
             shade = MfxCheckMenuItem(self),
             toolbar = MfxRadioMenuItem(self),
+            toolbar_size = MfxRadioMenuItem(self),
             toolbar_relief = MfxRadioMenuItem(self),
             statusbar = MfxCheckMenuItem(self),
         )
@@ -104,32 +117,41 @@ class PysolMenubarActions:
         if game is None:
             return
         assert self.app is game.app
+        tkopt, opt = self.tkopt, self.app.opt
         # set state of the menu items
-        self.tkopt.gameid.set(self.game.id)
-        self.tkopt.gameid_popular.set(self.game.id)
-        self.tkopt.confirm.set(self.app.opt.confirm)
-        self.tkopt.autofaceup.set(self.app.opt.autofaceup)
-        self.tkopt.autodrop.set(self.app.opt.autodrop)
-        self.tkopt.autodeal.set(self.app.opt.autodeal)
-        self.tkopt.quickplay.set(self.app.opt.quickplay)
-        self.tkopt.undo.set(self.app.opt.undo)
-        self.tkopt.hint.set(self.app.opt.hint)
-        self.tkopt.highlight_piles.set(self.app.opt.highlight_piles)
-        self.tkopt.highlight_cards.set(self.app.opt.highlight_cards)
-        self.tkopt.highlight_samerank.set(self.app.opt.highlight_samerank)
-        self.tkopt.sound.set(self.app.opt.sound)
-        self.tkopt.cardset.set(self.app.cardset.index)
-        self.tkopt.cardback.set(self.app.images.getCardbackIndex())
-        self.tkopt.tabletile.set(self.app.tabletile_index),
-        self.tkopt.animations.set(self.app.opt.animations)
-        self.tkopt.shadow.set(self.app.opt.shadow)
-        self.tkopt.shade.set(self.app.opt.shade)
-        self.tkopt.toolbar.set(self.app.opt.toolbar)
-        self.tkopt.toolbar_relief.set(self.app.opt.toolbar_relief)
-        self.tkopt.statusbar.set(self.app.opt.statusbar)
+        tkopt.gameid.set(game.id)
+        tkopt.gameid_popular.set(game.id)
+        tkopt.comment.set(bool(game.gsaveinfo.comment))
+        tkopt.autofaceup.set(opt.autofaceup)
+        tkopt.autodrop.set(opt.autodrop)
+        tkopt.autodeal.set(opt.autodeal)
+        tkopt.quickplay.set(opt.quickplay)
+        tkopt.undo.set(opt.undo)
+        tkopt.hint.set(opt.hint)
+        tkopt.bookmarks.set(opt.bookmarks)
+        tkopt.highlight_piles.set(opt.highlight_piles)
+        tkopt.highlight_cards.set(opt.highlight_cards)
+        tkopt.highlight_samerank.set(opt.highlight_samerank)
+        tkopt.sound.set(opt.sound)
+        tkopt.cardback.set(self.app.cardset.backindex)
+        tkopt.tabletile.set(self.app.tabletile_index),
+        tkopt.animations.set(opt.animations)
+        tkopt.shadow.set(opt.shadow)
+        tkopt.shade.set(opt.shade)
+        tkopt.toolbar.set(opt.toolbar)
+        tkopt.toolbar_size.set(opt.toolbar_size)
+        tkopt.toolbar_relief.set(opt.toolbar_relief)
+        tkopt.statusbar.set(opt.statusbar)
+
+    # will get called after connectGame()
+    def updateRecentGamesMenu(self, gameids):
+        pass
+
+    def updateBookmarkMenuState(self):
+        pass
 
     # will get called after a new cardset has been loaded
-    def updateBackgroundImages(self):
+    def updateBackgroundImagesMenu(self):
         pass
 
 
@@ -143,9 +165,9 @@ class PysolMenubarActions:
     def _cancelDrag(self):
         return self.game is None or self.game._cancelDrag()
 
-    def changed(self):
+    def changed(self, *args, **kw):
         assert self.game is not None
-        return self.game.changed()
+        return apply(self.game.changed, args, kw)
 
 
     #
@@ -158,22 +180,30 @@ class PysolMenubarActions:
     def setToolbarState(self, state, path):
         raise SubclassResponsibility
 
+    def _clearMenuState(self):
+        ms = self.menustate
+        for k, v in ms.__dict__.items():
+            if type(v) is types.ListType:
+                ms.__dict__[k] = [0] * len(v)
+            else:
+                ms.__dict__[k] = 0
+
     # update self.menustate for menu items and toolbar
     def _updateMenuState(self):
+        self._clearMenuState()
         game = self.game
         assert game is not None
         opt = self.app.opt
         ms = self.menustate
-        for k in ms.__dict__.keys():
-            ms.__dict__[k] = 0
         # 0 = DISABLED, 1 = ENABLED
         ms.save_as = game.canSaveGame()
+        ms.hold_and_quit = ms.save_as
         if game.filename and ms.save_as:
             ms.save = 1
         if opt.undo:
-            if game.moves.index > 0:
+            if game.canUndo() and game.moves.index > 0:
                 ms.undo = 1
-            if game.moves.index < len(game.moves.history):
+            if game.canRedo() and game.moves.index < len(game.moves.history):
                 ms.redo = 1
         if game.moves.index > 0:
             ms.restart = 1
@@ -200,18 +230,19 @@ class PysolMenubarActions:
             ms.rules = 1
 
     # update menu items and toolbar
-    def updateMenus(self):
+    def _updateMenus(self):
         if self.game is None:
             return
-        self._updateMenuState()
         ms = self.menustate
         # File menu
         self.setMenuState(ms.save, "file.save")
         self.setMenuState(ms.save_as, "file.saveas")
+        self.setMenuState(ms.hold_and_quit, "file.holdandquit")
         # Edit menu
         self.setMenuState(ms.undo, "edit.undo")
         self.setMenuState(ms.redo, "edit.redo")
         self.setMenuState(ms.redo, "edit.redoall")
+        self.updateBookmarkMenuState()
         self.setMenuState(ms.restart, "edit.restartgame")
         # Game menu
         self.setMenuState(ms.deal, "game.dealcards")
@@ -235,22 +266,40 @@ class PysolMenubarActions:
         self.setToolbarState(ms.redo, "redo")
         self.setToolbarState(ms.autodrop, "autodrop")
         self.setToolbarState(ms.rules, "rules")
+        #
+        self.tkopt.comment.set(bool(self.game.gsaveinfo.comment))
+
+    # update menu items and toolbar
+    def updateMenus(self):
+        if self.game is None:
+            return
+        self._updateMenuState()
+        self._updateMenus()
+
+    # disable menu items and toolbar
+    def disableMenus(self):
+        if self.game is None:
+            return
+        self._clearMenuState()
+        self._updateMenus()
 
 
     #
-    # menu actions
+    # File menu
     #
 
     def mNewGame(self, *args):
         if self._cancelDrag(): return
         if self.changed():
             if not self.game.areYouSure("New game"): return
-        if self.app.nextgame.cardset_index == self.app.cardset.index:
+        if self.game.nextGameFlags(self.game.id) == 0:
+            self.game.endGame()
             self.game.newGame()
         else:
-            self.game._quitGame(self.game.id)
+            self.game.endGame()
+            self.game.quitGame(self.game.id)
 
-    def _mSelectGame(self, id):
+    def _mSelectGame(self, id, random=None):
         if self._cancelDrag(): return
         if self.game.id == id:
             return
@@ -260,9 +309,8 @@ class PysolMenubarActions:
                 self.tkopt.gameid.set(self.game.id)
                 self.tkopt.gameid_popular.set(self.game.id)
                 return
-        self.tkopt.gameid.set(id)
-        self.tkopt.gameid_popular.set(id)
-        self.game._quitGame(id)
+        self.game.endGame()
+        self.game.quitGame(id, random=random)
 
     def mSelectGame(self, *args):
         self._mSelectGame(self.tkopt.gameid.get())
@@ -270,78 +318,97 @@ class PysolMenubarActions:
     def mSelectGamePopular(self, *args):
         self._mSelectGame(self.tkopt.gameid_popular.get())
 
-    def _mNewGameBySeed(self, seed):
+    def _mNewGameBySeed(self, seed, origin):
         try:
             id, random = constructRandom(seed)
             if id is None:
                 id = self.game.id
             if random is None:
                 return
+            if not self.app.getGameInfo(id):
+                raise ValueError
         except (ValueError, TypeError), ex:
             d = MfxDialog(self.top, title="Invalid game number",
                           text="Invalid game number\n" + str(seed),
                           bitmap="error")
             return
-        if id == self.game.id and self.app.nextgame.cardset_index == self.app.cardset.index:
+        f = self.game.nextGameFlags(id, random)
+        if f & 17 == 0:
+            return
+        random.origin = origin
+        if f & 15 == 0:
+            self.game.endGame()
             self.game.newGame(random=random)
         else:
-            self.game._quitGame(id, random=random)
+            self.game.endGame()
+            self.game.quitGame(id, random=random)
 
     def mNewGameWithNextId(self, *args):
+        if self._cancelDrag(): return
         if self.changed():
             if not self.game.areYouSure("Select next game number"): return
         r = self.game.random
         seed = r.increaseSeed(r.initial_seed)
         seed = r.str(seed)
-        self._mNewGameBySeed(seed)
+        self._mNewGameBySeed(seed, self.game.random.ORIGIN_NEXT_GAME)
 
     def mSelectGameById(self, *args):
         if self._cancelDrag(): return
-        id, f = None, self.game.getFullId(format=1)
+        id, f = None, self.game.getGameNumber(format=0)
         d = MfxSimpleEntry(self.top, "Select new game number", "\n\nEnter new game number", f,
-                           strings=("Ok", "Next game", "Cancel"), default=0,
+                           strings=("OK", "Next number", "Cancel"), default=0,
                            e_width=25)
         if d.status != 0: return
-        if d.num == 2: return
-        if d.num == 1:
+        if d.button == 2: return
+        if d.button == 1:
             self.mNewGameWithNextId()
             return
         if self.changed():
             if not self.game.areYouSure("Select new game number"): return
-        self._mNewGameBySeed(d.value)
+        self._mNewGameBySeed(d.value, self.game.random.ORIGIN_SELECTED)
 
     def mSelectRandomGame(self, *args):
         if self._cancelDrag(): return
         if self.changed():
             if not self.game.areYouSure("Select random game"): return
-        while 1:
-            id = self.app.getRandomGameId()
-            if 1 and id == self.game.id:        # force change of game
+        for i in range(1000):       # just in case, don't loop forever
+            gi = self.app.getGameInfo(self.app.getRandomGameId())
+            if gi is None:
+                continue
+            if 1 and gi.id == self.game.id:
+                # force change of game
+                continue
+            if 1 and gi.category != self.game.gameinfo.category:
+                # don't change game category
                 continue
             break
-        self.game._quitGame(id)
+        if gi and gi.id != self.game.id:
+            self.game.endGame()
+            self.game.quitGame(gi.id)
 
     def _mSelectNextGameFromList(self, gl, step):
         if self._cancelDrag(): return
         id = self.game.id
+        gl = list(gl)
         if len(gl) < 2 or not id in gl:
             return
         if self.changed():
             if not self.game.areYouSure("Select next game"): return
         index = (gl.index(id) + step) % len(gl)
-        self.game._quitGame(gl[index])
+        self.game.endGame()
+        self.game.quitGame(gl[index])
 
     def mSelectNextGameById(self, *args):
-        self._mSelectNextGameFromList(self.app.getGamesIdSortedById(), 1)
+        self._mSelectNextGameFromList(self.app.gdb.getGamesIdSortedById(), 1)
 
     def mSelectPrevGameById(self, *args):
-        self._mSelectNextGameFromList(self.app.getGamesIdSortedById(), -1)
+        self._mSelectNextGameFromList(self.app.gdb.getGamesIdSortedById(), -1)
 
     def mSelectNextGameByName(self, *args):
-        self._mSelectNextGameFromList(self.app.getGamesIdSortedByName(), 1)
+        self._mSelectNextGameFromList(self.app.gdb.getGamesIdSortedByName(), 1)
 
     def mSelectPrevGameByName(self, *args):
-        self._mSelectNextGameFromList(self.app.getGamesIdSortedByName(), -1)
+        self._mSelectNextGameFromList(self.app.gdb.getGamesIdSortedByName(), -1)
 
     def mSave(self, *args):
         if self._cancelDrag(): return
@@ -351,39 +418,81 @@ class PysolMenubarActions:
             else:
                 self.mSaveAs()
 
+    def mHoldAndQuit(self, *args):
+        if self._cancelDrag(): return
+        self.game.endGame(holdgame=1)
+        self.game.quitGame(holdgame=1)
+
     def mQuit(self, *args):
         if self._cancelDrag(): return
         if self.changed():
             if not self.game.areYouSure("Quit PySol"): return
-        self.game._quitGame()
+        self.game.endGame()
+        self.game.quitGame()
+
+
+    #
+    # Edit menu
+    #
 
     def mUndo(self, *args):
         if self._cancelDrag(): return
-        if self.app.opt.undo:
+        if self.menustate.undo:
+            self.game.playSample("undo")
             self.game.undo()
 
     def mRedo(self, *args):
         if self._cancelDrag(): return
-        if self.app.opt.undo:
+        if self.menustate.redo:
+            self.game.playSample("redo")
             self.game.redo()
             self.game.checkForWin()
 
     def mRedoAll(self, *args):
         if self._cancelDrag(): return
-        if self.app.opt.undo:
+        if self.menustate.redo:
+            self.game.playSample("redo", loop=1)
             while self.game.moves.index < len(self.game.moves.history):
                 self.game.redo()
                 if self.game.checkForWin():
                     break
+            self.game.stopSamples()
+
+    def mSetBookmark(self, n, confirm=1):
+        if self._cancelDrag(): return
+        if not self.app.opt.bookmarks: return
+        if not (0 <= n <= 8): return
+        self.game.setBookmark(n, confirm=confirm)
+        self.game.updateMenus()
+
+    def mGotoBookmark(self, n, confirm=-1):
+        if self._cancelDrag(): return
+        if not self.app.opt.bookmarks: return
+        if not (0 <= n <= 8): return
+        self.game.gotoBookmark(n, confirm=confirm)
+        self.game.updateMenus()
+
+    def mClearBookmarks(self, *args):
+        if self._cancelDrag(): return
+        if not self.app.opt.bookmarks: return
+        if not self.game.gsaveinfo.bookmarks: return
+        if not self.game.areYouSure("Clear bookmarks","Clear all bookmarks ?"): return
+        self.game.gsaveinfo.bookmarks = {}
+        self.game.updateMenus()
 
     def mRestart(self, *args):
         if self._cancelDrag(): return
         if self.game.moves.index == 0:
             return
-        if self.changed():
+        if self.changed(restart=1):
             if not self.game.areYouSure("Restart game","Restart this game ?"): return
         self.game.restartGame()
 
+
+    #
+    # Game menu
+    #
+
     def mDeal(self, *args):
         if self._cancelDrag(): return
         self.game.dealCards()
@@ -397,120 +506,132 @@ class PysolMenubarActions:
         self.game.autoPlay(autofaceup=1, autodrop=1)
 
     def mStatus(self, *args):
-        game, stats = self.game, self.game.stats
         if self._cancelDrag(): return
-        w = ""
-        if game.s.talon:
-            w = w + "\nRedeals: " + str(game.s.talon.round - 1)
-            w = w + "\nCards in Talon: " + str(len(game.s.talon.cards))
-        if game.s.waste:
-            w = w + "\nCards in Waste: " + str(len(game.s.waste.cards))
-        if game.s.foundations:
-            n = 0
-            for s in game.s.foundations:
-                n = n + len(s.cards)
-            w = w + "\nCards in Foundations: " + str(n)
-        d = MfxDialog(self.top,title="Game status",
-                   text=game.getTitleName() + "\n" +
-                        game.getFullId(format=2) + "\n" +
-                        "Playing time: " + game.getTime() + "\n\n" +
-                        "Moves: " + str(game.moves.index) + "\n" +
-                        "Undo moves: " + str(stats.undo_moves) + "\n" +
-                        "Demo moves: " + str(stats.demo_moves) + "\n\n" +
-###                        "Total player moves: " + str(stats.player_moves) + "\n" +
-###                        "Total moves in this game: " + str(stats.total_moves) + "\n" +
-                        "Hints: " + str(stats.hints) + "\n" +
-                        "Highlight piles: " + str(stats.highlight_piles) + "\n" +
-                        "Highlight cards: " + str(stats.highlight_cards) + "\n" +
-                        "Highlight same rank: " + str(stats.highlight_samerank) + "\n" +
-                        w,
-                   strings=("Ok", "Stats...", "Log...", "Demo..."),
-                   padx=30, separatorwidth=2)
-        if d.status == 0:
-            if d.num == 1:
-                self.mPlayerStats()
-            if d.num == 2:
-                self.mPlayerLog()
-            elif d.num == 3:
-                self.mDemoStats()
+        self.mPlayerStats(mode=100)
 
-    def _mHandleStats(self, player, header, title, num):
-        n = self.game.getTitleName()
-        # print to file
-        if num == 3:
-            file = None
-            filename, text = "stats.txt", "Your statistics"
-            if player is None:
-                filename, text = "stats_demo.txt", "Demo statistics"
-            filename = os.path.join(self.app.dn.config, filename)
-            filename = os.path.normpath(filename)
-            try:
-                file = open(filename, "a")
-                a = PysolStatsWriter(self.app)
-                writer = a.FileWriter(file)
-                a.writeFullStats(writer, player, header, title)
-                destruct(a)
-            except EnvError, ex:
-                if file: file.close()
-                d = MfxExceptionDialog(self.top, ex,
-                              text="Error while writing to file")
+    def mEditGameComment(self, *args):
+        if self._cancelDrag(): return
+        game, gi = self.game, self.game.gameinfo
+        t = " " + game.getGameNumber(format=1)
+        cc = "Comments for " + gi.name + t + ":\n\n"
+        c = game.gsaveinfo.comment or cc
+        d = EditTextDialog(game.top, "Comments for"+t, text=c)
+        if d.status == 0 and d.button == 0:
+            if string.strip(d.text) == string.strip(cc):
+                game.gsaveinfo.comment = ""
             else:
-                if file: file.close()
-                d = MfxDialog(self.top, title="PySol Info", bitmap="info",
-                              text=text + " were appended to\n" + filename)
-            return
-        # reset
-        if player is not None:
-            if num == 1:
-                if self.game.areYouSure("Reset player statistics",
-                                        "Reset your statistics for\n" + n + " ?",
-                                        confirm=1, default=1):
-                    self.app.stats.resetStats(player, self.game.id)
-            elif num == 2:
-                if self.game.areYouSure("Reset player statistics",
-                                        "Reset ALL statistics for\nplayer " + player + " ?",
-                                        confirm=1, default=1):
-                    self.app.stats.resetStats(player, 0)
-            self.game.updateStatus(stats=self.app.stats.getStats(player, self.game.id))
+                game.gsaveinfo.comment = string.rstrip(d.text)
+        self.tkopt.comment.set(bool(game.gsaveinfo.comment))
+
+
+    #
+    # Game menu - statistics
+    #
+
+    def _mStatsSave(self, player, header, filename, write_method):
+        file = None
+        if player is None:
+            text = "Demo statistics"
+            filename = filename + "_demo"
         else:
-            if num == 1:
-                if self.game.areYouSure("Reset demo statistics",
-                                        "Reset demo statistics for\n" + n + " ?",
-                                         confirm=1, default=1):
-                    self.app.stats.resetStats(player, self.game.id)
-            elif num == 2:
-                if self.game.areYouSure("Reset demo statistics",
-                                        "Reset ALL demo statistics ?",
+            text = "Your statistics"
+        filename = os.path.join(self.app.dn.config, filename + ".txt")
+        filename = os.path.normpath(filename)
+        try:
+            file = open(filename, "a")
+            a = PysolStatsFormatter(self.app)
+            writer = a.FileWriter(file)
+            apply(write_method, (a, writer, player, header))
+            destruct(a)
+        except EnvError, ex:
+            if file: file.close()
+            d = MfxExceptionDialog(self.top, ex,
+                              text="Error while writing to file")
+        else:
+            if file: file.close()
+            d = MfxDialog(self.top, title=PACKAGE+" Info", bitmap="info",
+                          text=text + " were appended to\n\n" + filename)
+
+
+    def mPlayerStats(self, *args, **kw):
+        mode = kw.get("mode", 101)
+        demo = 0
+        while mode > 0:
+            if mode > 1000:
+                demo = not demo
+                mode = mode % 1000
+            #
+            d = Struct(status=-1, button=-1)
+            if demo:
+                player = None
+                p0, p1, p2 = PACKAGE+" Demo", PACKAGE+" Demo ", ""
+            else:
+                player = self.app.opt.player
+                p0, p1, p2 = player, "", " for " + player
+            n = self.game.gameinfo.short_name
+            #
+            if mode == 100:
+                d = Status_StatsDialog(self.top, game=self.game)
+            elif mode == 101:
+                header = p1 + "Statistics for " + n
+                d = SingleGame_StatsDialog(self.top, header, self.app, player, gameid=self.game.id)
+            elif mode == 102:
+                header = p1 + "Statistics" + p2
+                d = AllGames_StatsDialog(self.top, header, self.app, player)
+            elif mode == 103:
+                header = p1 + "Full log" + p2
+                d = FullLog_StatsDialog(self.top, header, self.app, player)
+            elif mode == 104:
+                header = p1 + "Session log" + p2
+                d = SessionLog_StatsDialog(self.top, header, self.app, player)
+            elif mode == 202:
+                # print stats to file
+                header = "Statistics for " + p0
+                write_method = PysolStatsFormatter.writeStats
+                self._mStatsSave(player, header, "stats", write_method)
+            elif mode == 203:
+                # print full log to file
+                header = "Full log for " + p0
+                write_method = PysolStatsFormatter.writeFullLog
+                self._mStatsSave(player, header, "log", write_method)
+            elif mode == 204:
+                # print session log to file
+                header = "Session log for " + p0
+                write_method = PysolStatsFormatter.writeSessionLog
+                self._mStatsSave(player, header, "log", write_method)
+            elif mode == 301:
+                # reset all player stats
+                if self.game.areYouSure("Reset all statistics",
+                                        "Reset ALL statistics and logs for player\n" + p0 + " ?",
                                         confirm=1, default=1):
                     self.app.stats.resetStats(player, 0)
+                    self.game.updateStatus(stats=self.app.stats.getStats(self.app.opt.player, self.game.id))
+            elif mode == 302:
+                # reset player stats for current game
+                if self.game.areYouSure("Reset game statistics",
+                                        "Reset statistics and logs for player\n" + p0 + "\nand game\n" + n +" ?",
+                                        confirm=1, default=1):
+                    self.app.stats.resetStats(player, self.game.id)
+                    self.game.updateStatus(stats=self.app.stats.getStats(self.app.opt.player, self.game.id))
+            elif mode == 401:
+                # start a new game with a gameid
+                ## TODO
+                pass
+            elif mode == 402:
+                # start a new game with a gameid / gamenumber
+                ## TODO
+                pass
+            else:
+                print "stats problem:", mode, demo, player
+                pass
+            if d.status != 0:
+                break
+            mode = d.button
 
-    # default implementation uses a simple dialog with a fixed size font
-    def _mDoStats(self, player, header, title):
-        a = PysolStatsWriter(self.app)
-        writer = a.StringWriter()
-        a.writeStats(writer, player, header)
-        text = writer.text
-        destruct(a)
-        font = buttonfont = getFont("fixed")
-        d = MfxDialog(self.top, title=title,
-                      text=text, justify="left",
-                      font=font, buttonfont=buttonfont,
-                      strings=("Ok", "Reset game", "Reset all", "Print to file"),
-                      padx=40, separatorwidth=2)
-        if d.status == 0:
-            self._mHandleStats(player, header, title, d.num)
 
-    def mPlayerStats(self, *args):
-        if self._cancelDrag(): return
-        player = self.app.opt.player
-        self._mDoStats(player, player, title="Statistics for " + player)
-
-    def mDemoStats(self, *args):
-        if self._cancelDrag(): return
-        self._mDoStats(None, "Demo games", title="PySol demo statistics")
-
-    def mPlayerLog(self, *args):
-        pass
+    #
+    # Assist menu
+    #
 
     def mHint(self, *args):
         if self._cancelDrag(): return
@@ -548,21 +669,25 @@ class PysolMenubarActions:
         ##self.app.demo_counter = 0
         self.game.startDemo(mixed=mixed)
 
-    def mOptPlayerName(self, *args):
+
+    #
+    # Options menu
+    #
+
+    def mOptPlayerOptions(self, *args):
         if self._cancelDrag(): return
-        d = MfxSimpleEntry(self.top, "Set player name", "\nPlease enter your name",
-                           self.app.opt.player, e_width=20)
-        if d.status == 0:
-            n = string.strip(d.value)
+        d = PlayerOptionsDialog(self.top, "Set player options", self.app)
+        if d.status == 0 and d.button == 0:
+            self.app.opt.confirm = bool(d.confirm)
+            self.app.opt.update_player_stats = bool(d.update_stats)
+            self.app.opt.win_animation = bool(d.win_animation)
+            ##n = string.strip(d.player)
+            n = string.strip(d.player[:30])
             if 0 < len(n) <= 30:
                 self.app.opt.player = n
                 self.game.updateStatus(player=self.app.opt.player)
                 self.game.updateStatus(stats=self.app.stats.getStats(self.app.opt.player, self.game.id))
 
-    def mOptConfirm(self, *args):
-        if self._cancelDrag(): return
-        self.app.opt.confirm = self.tkopt.confirm.get()
-
     def mOptAutoFaceUp(self, *args):
         if self._cancelDrag(): return
         self.app.opt.autofaceup = self.tkopt.autofaceup.get()
@@ -590,6 +715,11 @@ class PysolMenubarActions:
         self.app.opt.undo = self.tkopt.undo.get()
         self.game.updateMenus()
 
+    def mOptEnableBookmarks(self, *args):
+        if self._cancelDrag(): return
+        self.app.opt.bookmarks = self.tkopt.bookmarks.get()
+        self.game.updateMenus()
+
     def mOptEnableHint(self, *args):
         if self._cancelDrag(): return
         self.app.opt.hint = self.tkopt.hint.get()
@@ -613,10 +743,13 @@ class PysolMenubarActions:
     def mOptSound(self, *args):
         if self._cancelDrag(): return
         self.app.opt.sound = self.tkopt.sound.get()
+        if not self.app.opt.sound:
+            self.app.audio.stopAll()
 
-    def mOptCardset(self, *args):
+    def mOptSoundDialog(self, *args):
         if self._cancelDrag(): return
-        self.app.nextgame.cardset_index = self.tkopt.cardset.get()
+        d = SoundOptionsDialog(self.top, "Sound settings", self.app)
+        self.tkopt.sound.set(self.app.opt.sound)
 
     def mOptAnimations(self, *args):
         if self._cancelDrag(): return
@@ -634,19 +767,19 @@ class PysolMenubarActions:
         if self._cancelDrag(): return
         self.app.opt.irregular_piles = self.tkopt.irregular_piles.get()
 
-    def mOptDemoSpeed(self, *args):
+    def mOptDemoOptions(self, *args):
         if self._cancelDrag(): return
-        d = MfxSimpleSlider(self.top, "Set demo speed", "Set demo delay in seconds",
-                            self.app.opt.demo_sleep, 0.2, 9.9, 0.1)
-        if d.status == 0:
-            self.app.opt.demo_sleep = d.value
+        d = DemoOptionsDialog(self.top, "Set demo options", self.app)
+        if d.status == 0 and d.button == 0:
+            self.app.opt.demo_logo = d.demo_logo
+            self.app.opt.demo_score = d.demo_score
+            self.app.opt.demo_sleep = d.demo_sleep
 
-    def mOptHintSpeed(self, *args):
+    def mOptHintOptions(self, *args):
         if self._cancelDrag(): return
-        d = MfxSimpleSlider(self.top, "Set hint speed", "Set hint delay in seconds",
-                            self.app.opt.hint_sleep, 0.2, 9.9, 0.1)
-        if d.status == 0:
-            self.app.opt.hint_sleep = d.value
+        d = HintOptionsDialog(self.top, "Set hint options", self.app)
+        if d.status == 0 and d.button == 0:
+            self.app.opt.hint_sleep = d.hint_sleep
 
     def mOptSave(self, *args):
         if self._cancelDrag(): return
@@ -657,17 +790,21 @@ class PysolMenubarActions:
                                    text="Error while saving options")
         else:
             # tell the player where their config files reside
-            d = MfxDialog(self.top, title="PySol Info",
+            d = MfxDialog(self.top, title=PACKAGE+" Info",
                           text="Options were saved to\n\n" + self.app.fn.opt,
                           bitmap="info")
 
+    #
+    # Help menu
+    #
+
     def mHelp(self, *args):
         if self._cancelDrag(): return
         helpHTML(self.app, "index.html", "html")
 
     def mHelpHowToPlay(self, *args):
         if self._cancelDrag(): return
-        helpHTML(self.app, "usage.html", "html")
+        helpHTML(self.app, "howtoplay.html", "html")
 
     def mHelpRules(self, *args):
         if self._cancelDrag(): return
@@ -679,11 +816,52 @@ class PysolMenubarActions:
 
     def mHelpLicense(self, *args):
         if self._cancelDrag(): return
-        helpHTML(self.app, "gpl.html", "html")
+        helpHTML(self.app, "license.html", "html")
+
+    def mHelpNews(self, *args):
+        if self._cancelDrag(): return
+        helpHTML(self.app, "news.html", "html")
+
+    def mHelpWebSite(self, *args):
+        openURL(PACKAGE_URL)
 
     def mHelpAbout(self, *args):
         if self._cancelDrag(): return
-        helpAbout(self.app)
+        if self.app and bundle & 4:
+            helpAboutSimple(self.app)
+        else:
+            helpAbout(self.app)
+
+
+    #
+    # misc
+    #
+
+    def mScreenshot(self, *args):
+        if self._cancelDrag(): return
+        f = os.path.join(self.app.dn.config, "screenshots")
+        if not os.path.isdir(f):
+            return
+        f = os.path.join(f, self.app.getGameSaveName(self.game.id))
+        i = 1
+        while 1:
+            fn = f + ("-%d.ppm" % i)
+            if not os.path.isfile(fn):
+                break
+            i = i + 1
+            if i >= 10000:      # give up
+                return
+        self.top.screenshot(fn)
+
+    def mPlayNextMusic(self, *args):
+        if self._cancelDrag(): return
+        if self.app.audio and self.app.opt.sound_music_volume > 0:
+            self.app.audio.playNextMusic()
+            if 1 and self.app.debug:
+                index = self.app.audio.getMusicInfo()
+                music = self.app.music_manager.get(index)
+                if music:
+                    print "playing music:", music.filename
 
 
 # /***********************************************************************
@@ -766,8 +944,8 @@ class PysolToolbarActions:
             self.menubar.mQuit()
         return 1
 
-    def mOptPlayerName(self, *args):
+    def mOptPlayerOptions(self, *args):
         if not self._busy():
-            self.menubar.mOptPlayerName()
+            self.menubar.mOptPlayerOptions()
         return 1
 
