@@ -4,6 +4,7 @@
 ##
 ## PySol -- a Python Solitaire game
 ##
+## Copyright (C) 2000 Markus Franz Xaver Johannes Oberhumer
 ## Copyright (C) 1999 Markus Franz Xaver Johannes Oberhumer
 ## Copyright (C) 1998 Markus Franz Xaver Johannes Oberhumer
 ##
@@ -23,18 +24,19 @@
 ## 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 ##
 ## Markus F.X.J. Oberhumer
-## <markus.oberhumer@jk.uni-linz.ac.at>
-## http://wildsau.idv.uni-linz.ac.at/mfx/pysol.html
+## <markus@oberhumer.com>
+## http://www.oberhumer.com/pysol
 ##
 ##---------------------------------------------------------------------------##
 
 
 # imports
-import sys
+import string, struct, sys
 
 # PySol imports
-from mfxutil import Struct                                          #bundle#
-from util import KING                                               #bundle#
+from mfxtools import *
+from mfxutil import Struct, destruct
+from util import KING
 
 
 # /***********************************************************************
@@ -110,9 +112,7 @@ class AbstractHint(HintInterface):
             # copy class identity
             self.__class__ = stack.__class__
             # copy model data (reference copy)
-            self.id = stack.id
-            self.game = stack.game
-            self.cap = stack.cap
+            stack.copyModel(self)
             # set new cards (shallow copy of the card list)
             self.cards = stackcards[:]
 
@@ -124,7 +124,7 @@ class AbstractHint(HintInterface):
     def __destructClones(self):
         for s in self.__clones:
             s.__class__ = self.AClonedStack     # restore orignal class
-            del s.game, s.cap, s.cards
+            destruct(s)
         self.__clones = []
 
     # When computing hints for level 0, the scores are flattened
@@ -203,17 +203,17 @@ class AbstractHint(HintInterface):
 
     # we move the pile if it is accepted by the target stack
     def _defaultShallMovePile(self, from_stack, to_stack, pile, rpile):
-        if from_stack is to_stack or not to_stack.acceptsPile(from_stack, pile):
+        if from_stack is to_stack or not to_stack.acceptsCards(from_stack, pile):
             return 0
         return 1
 
     # same, but check for loops
     def _cautiousShallMovePile(self, from_stack, to_stack, pile, rpile):
-        if from_stack is to_stack or not to_stack.acceptsPile(from_stack, pile):
+        if from_stack is to_stack or not to_stack.acceptsCards(from_stack, pile):
             return 0
         # now check for loops
         rr = self.ClonedStack(from_stack, stackcards=rpile)
-        if rr.acceptsPile(to_stack, pile):
+        if rr.acceptsCards(to_stack, pile):
             # the pile we are going to move could be moved back -
             # this is dangerous as we can create endless loops...
             return 0
@@ -221,12 +221,12 @@ class AbstractHint(HintInterface):
 
     # same, but only check for loops only when in demo mode
     def _cautiousDemoShallMovePile(self, from_stack, to_stack, pile, rpile):
-        if from_stack is to_stack or not to_stack.acceptsPile(from_stack, pile):
+        if from_stack is to_stack or not to_stack.acceptsCards(from_stack, pile):
             return 0
         if self.level >= 2:
             # now check for loops
             rr = self.ClonedStack(from_stack, stackcards=rpile)
-            if rr.acceptsPile(to_stack, pile):
+            if rr.acceptsCards(to_stack, pile):
                 # the pile we are going to move could be moved back -
                 # this is dangerous as we can create endless loops...
                 return 0
@@ -248,7 +248,7 @@ class AbstractHint(HintInterface):
         for card in cards:
             for s in stacks:
                 if s is not from_stack:
-                    if s.acceptsPile(from_stack, [card]):
+                    if s.acceptsCards(from_stack, [card]):
                         break
             else:
                 return 0
@@ -355,7 +355,7 @@ class DefaultHint(AbstractHint):
             self.bonus_color = self.RED
             return self.BONUS_CAN_DROP_ALL_CARDS + self.BONUS_CAN_CREATE_EMPTY_ROW
         # check if the cards below our pile are a whole row
-        if r.canMovePile(rpile):
+        if r.canMoveCards(rpile):
             # could we move the remaining pile ?
             for x in self.game.s.rows:
                 # note: we allow x == r here, because the pile
@@ -363,7 +363,7 @@ class DefaultHint(AbstractHint):
                 #       available in the next move
                 if x is t or not x.cards:
                     continue
-                if x.acceptsPile(r, rpile):
+                if x.acceptsCards(r, rpile):
                     # we can create an empty row in the next move
                     ##print "BONUS_CAN_CREATE_EMPTY_ROW", r, x, pile, rpile
                     self.bonus_color = self.BLUE
@@ -406,7 +406,14 @@ class DefaultHint(AbstractHint):
              return 93000, color
         pile = r.cards
         c = pile[-1]
-        if c.rank <= 1:
+        # compute distance to t.cap.base_rank - compare Stack.getRankDir()
+        if t.cap.base_rank < 0:
+            d = len(t.cards)
+        else:
+            d = (c.rank - t.cap.base_rank) % t.cap.mod
+            if d > t.cap.mod / 2:
+                d = d - t.cap.mod
+        if abs(d) <= 1:
              # drop Ace and 2 immediately
              score = 92000
         elif r in self.game.sg.talonstacks:
@@ -464,7 +471,7 @@ class DefaultHint(AbstractHint):
     def step010(self, dropstacks, rows):
         # for each stack
         for r in dropstacks:
-            # 3a) try if we can drop cards
+            # 1a) try if we can drop cards
             t, ncards = r.canDropCards(self.game.s.foundations)
             if t:
                 score, color = 0, None
@@ -472,7 +479,7 @@ class DefaultHint(AbstractHint):
                 self.addHint(score, ncards, r, t, color)
                 if score >= 90000:
                     break
-            # 3b) try if we can move cards to one of the RowStacks
+            # 1b) try if we can move cards to one of the RowStacks
             for pile in self.step010b_getPiles(r):
                 if pile:
                     self.step010_movePile(r, pile, rows)
@@ -542,9 +549,9 @@ class DefaultHint(AbstractHint):
                     c = di[0]
                     sub_pile = pile[di[3]+1 : ]
                     ##print "trying drop move", c, pile, sub_pile
-                    assert r.canMovePile(sub_pile)
+                    assert r.canMoveCards(sub_pile)
                     for t in rows:
-                        if t is r or not t.acceptsPile(r, sub_pile):
+                        if t is r or not t.acceptsCards(r, sub_pile):
                             continue
                         ##print "drop move", r, t, sub_pile
                         score = 40000
@@ -560,11 +567,11 @@ class DefaultHint(AbstractHint):
     def step030(self, foundations, rows, dropstacks):
         for s in foundations:
             card = s.getCard()
-            if not card or not s.canMovePile([card]):
+            if not card or not s.canMoveCards([card]):
                 continue
             # search a RowStack that would accept the card
             for t in rows:
-                if t is s or not t.acceptsPile(s, [card]):
+                if t is s or not t.acceptsCards(s, [card]):
                     continue
                 tt = self.ClonedStack(t, stackcards=t.cards+[card])
                 # search a Stack that would benefit from this card
@@ -574,12 +581,12 @@ class DefaultHint(AbstractHint):
                     pile = r.getPile()
                     if not pile:
                         continue
-                    if not tt.acceptsPile(r, pile):
+                    if not tt.acceptsCards(r, pile):
                         continue
                     # compute remaining pile in r
                     rpile = r.cards[ : (len(r.cards)-len(pile)) ]
                     rr = self.ClonedStack(r, stackcards=rpile)
-                    if rr.acceptsPile(t, pile):
+                    if rr.acceptsCards(t, pile):
                         # the pile we are going to move from r to t
                         # could be moved back from t ro r - this is
                         # dangerous as we can create loops...
@@ -599,16 +606,16 @@ class DefaultHint(AbstractHint):
             return
         for r in rows:
             card = r.getCard()
-            if not card or not r.canMovePile([card]):
+            if not card or not r.canMoveCards([card]):
                 continue
             pile = [card]
             # compute remaining pile in r
             rpile = r.cards[ : (len(r.cards)-len(pile)) ]
             rr = self.ClonedStack(r, stackcards=rpile)
             for t in reservestacks:
-                if t is r or not t.acceptsPile(r, pile):
+                if t is r or not t.acceptsCards(r, pile):
                     continue
-                if rr.acceptsPile(t, pile):
+                if rr.acceptsCards(t, pile):
                     # the pile we are going to move from r to t
                     # could be moved back from t ro r - this is
                     # dangerous as we can create loops...
@@ -633,6 +640,10 @@ class DefaultHint(AbstractHint):
 
 class CautiousDefaultHint(DefaultHint):
     shallMovePile = DefaultHint._cautiousShallMovePile
+    ##shallMovePile = DefaultHint._cautiousDemoShallMovePile
+
+    def _preferHighRankMoves(self):
+        return 1
 
 
 # /***********************************************************************
@@ -664,3 +675,79 @@ class GolfType_Hint(DefaultHint):
 class SpiderType_Hint(DefaultHint):
     pass
 
+
+
+# /***********************************************************************
+# //
+# ************************************************************************/
+
+#%ifndef BUNDLE
+pysolfreecellsolver = None
+try:
+    import pysolfreecellsolver
+except:
+    pysolfreecellsolver = None
+#%endif
+
+
+class FreeCellSolverWrapper:
+    class FreeCellSolver_Hint(AbstractHint):
+        def str1(self, card):
+            return "A23456789TJQK"[card.rank] + "CSHD"[card.suit]
+        def str2(self, card):
+            return "CSHD"[card.suit] + "-" + "A23456789TJQK"[card.rank]
+
+        def computeHints(self):
+            board = ""
+            #
+            b = ""
+            for s in self.game.s.foundations:
+                if s.cards:
+                    b = b + " " + self.str2(s.cards[-1])
+            if b:
+                board = board  + "Founds:" + b + "\n"
+            #
+            b = ""
+            for s in self.game.s.reserves:
+                if s.cards:
+                    b = b + " " + self.str1(s.cards[-1])
+                else:
+                    b = b + " -"
+            if b:
+                board = board  + "FC:" + b + "\n"
+            #
+            for s in self.game.s.rows:
+                b = ""
+                for c in s.cards:
+                    b = b + self.str1(c) + " "
+                board = board + string.strip(b) + "\n"
+            #
+            ##print board
+            solver = apply(pysolfreecellsolver.FCSolver, self.fcs_args)
+            try:
+                h = solver.solve(board)
+            except:
+                h = None
+            if h:
+                i = len(h)
+                assert i % 3 == 0
+                hint = None
+                map = self.game.s.foundations + self.game.s.rows + self.game.s.reserves
+                while i > 0:
+                    i = i - 3
+                    v = struct.unpack("<BBB", h[i:i+3])
+                    hint = (999999, 0, v[0], map[v[1]], map[v[2]], None, hint)
+                self.hints.append(hint)
+
+    def __init__(self, Fallback_Class, *args):
+        self.Fallback_Class = Fallback_Class
+        self.args = args
+
+    def __call__(self, game, level):
+        if pysolfreecellsolver is None:
+            hint = self.Fallback_Class(game, level)
+        else:
+            hint = self.FreeCellSolver_Hint(game, level)
+            hint.fcs_args = (game.gameinfo.decks, len(game.s.rows), len(game.s.reserves)) + self.args
+        return hint
+
