@@ -4,6 +4,8 @@
 ##
 ## PySol -- a Python Solitaire game
 ##
+## Copyright (C) 2001 Markus Franz Xaver Johannes Oberhumer
+## Copyright (C) 2000 Markus Franz Xaver Johannes Oberhumer
 ## Copyright (C) 1999 Markus Franz Xaver Johannes Oberhumer
 ## Copyright (C) 1998 Markus Franz Xaver Johannes Oberhumer
 ##
@@ -23,23 +25,27 @@
 ## 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 ##
 ## Markus F.X.J. Oberhumer
-## <markus.oberhumer@jk.uni-linz.ac.at>
-## http://wildsau.idv.uni-linz.ac.at/mfx/pysol.html
+## <markus@oberhumer.com>
+## http://www.oberhumer.com/pysol
 ##
 ##---------------------------------------------------------------------------##
 
 
 # imports
-import os, sys, Tkinter, Canvas
+import os, sys, re, string, types, Tkinter
+import traceback
 
 # PySol imports
-from mfxutil import destruct, kwdefault, KwStruct                   #bundle#
+from mfxutil import destruct, kwdefault, KwStruct
+from mfxutil import win32api
 
 # Toolkit imports
-from tkconst import EVENT_HANDLED, EVENT_PROPAGATE                  #bundle#
-from tkutil import getFont, after, after_cancel                     #bundle#
-from tkutil import bind, unbind_destroy                             #bundle#
-from tkutil import makeToplevel, setTransient                       #bundle#
+from tkconst import tkversion
+from tkconst import EVENT_HANDLED, EVENT_PROPAGATE
+from tkutil import getFont
+from tkutil import after, after_idle, after_cancel
+from tkutil import bind, unbind_destroy, makeImage
+from tkutil import makeToplevel, setTransient
 
 
 # /***********************************************************************
@@ -48,33 +54,34 @@ from tkutil import makeToplevel, setTransient                       #bundle#
 
 class _ToplevelDialog:
     img = None
-    def __init__(self, parent, title="", resizable=0):
+    def __init__(self, parent, title="", resizable=0, default=-1):
         self.parent = parent
         self.status = 0
+        self.button = default
         self.timer = None
         self.top = makeToplevel(parent, title=title)
         self.top.wm_resizable(resizable, resizable)
         bind(self.top, "WM_DELETE_WINDOW", self.wmDeleteWindow)
         if self.img is None:
-            self.img = (Tkinter.PhotoImage(data="""
+            _ToplevelDialog.img = (makeImage(dither=0, data="""
 R0lGODlhIAAgAPICAH8AAH9/f7+/v/8AAP///wAAAAAAAAAAACH5BAEAAAIALAAAAAAgACAAAAPZ
 KLos8PC1SRccOGNYO9AgCHTNE57ZQwoA6mZAB7z0AFRA/QLBBISEFyEECDAAIYISpSSEAAwAqKkM
 UQkgQGABCF011xAgEBAAUFfMFQUICACu69UFCAgAL7nyBSADaHI0AGQANHI0AGQAL3oELwBkAC6N
 Si4AZAAoVxhXKABkAgAhVxpXIQBkAQIAIFchVyAAqQIAIU0oTSEAZAsAIQQvBCEAqQwAOi4AqQEM
 AMgoAKkTAM8gAKkVD9UPyx0BAMgAywErARAoEOMrC2QREeMB6w3w9AHyFfUrCQA7"""),
-Tkinter.PhotoImage(data="""
+makeImage(dither=0, data="""
 R0lGODlhIAAgAPIDAAAAAAAA/39/f7+/v////wAAAAAAAAAAACH5BAEAAAMALAAAAAAgACAAAAPU
 OLo88vC1SRccJOsBa32DJo7D4zHCMK7jIJyDwA5BHQws4AksYds9AEXQI/xqRUBDMCgOfoPiAMAQ
 FK8sgEAhGGC/mgFAMBBcjzUsQDAQfI9YAOR7xAIg3yMWAPkesQAQA1hHVwMADwMCWEdXABADAgNX
 R0UDABAKAj1oNiwAEAwAYEUAEQ0ApCsAERQAA6oEAwARHgAAA18DtxEwArcAAysDwBECMAoQwAMa
 AMYPyAwRABoAxtEVAgAaAA/YyAAZAA/fyAAEAA/lyLcP68gQ78gPMAkAOw=="""),
-Tkinter.PhotoImage(data="""
+makeImage(dither=0, data="""
 R0lGODlhIAAgAPIDAAAAAAAA/39/f7+/v////wAAAAAAAAAAACH5BAEAAAMALAAAAAAgACAAAAPp
 OLo88vC1SRccJOsBa32DJo7D4zHCMK7jIJyDwM4j4AnsEOz8wAIUAWsQGGR4gQEL0BAMWDsRkjUA
 MAQzZAY5AwgUgsEMSRggZwOAYCCgDXYDApJGAAgGArpmwNMDIHoZPIEAEIEEOwOBABCHO4cAEAOH
 hwMADwMCegM7A3oAEAMCA3RIdAMAEAoCdEh0ABAMADQDOwM0ABENAJQrABEUAAO9BAMAER4AAAOB
 A8oRMALKAAMrA9MRAjAKENMDGgDZD9sMEQAaANnkFQIAGgAP69sAGQAP8tsABAAP+NvKD/xtgyBw
-2wMYCQAAOw=="""), Tkinter.PhotoImage(data="""
+2wMYCQAAOw=="""), makeImage(dither=0, data="""
 R0lGODlhIAAgAPIDAAAAAH9/AH9/f7+/v///AAAAAAAAAAAAACH5BAEAAAMALAAAAAAgACAAAAPu
 OLq88S3KqQIhA9Atw70DIHDkEHwXIAglFaDXAKytFMAXsNZNgF8DwIq3CPwugBVxEDheBoAVMfAZ
 AK6DD2DFC6Cu4M8AsKoFUOArCrBqBWBpAGoAWJUCsDgMsCIFcHEwAwArHAE4AWABOAArGwE/A2AD
@@ -84,14 +91,14 @@ oEEBEQ4qJKggAQA7"""))
 
     def mainloop(self, focus=None, timeout=0):
         bind(self.top, "<Escape>", self.mCancel)
+        if focus is not None:
+            focus.focus()
         setTransient(self.top, self.parent)
         try:
             self.top.grab_set()
         except Tkinter.TclError:
             print "*** Grab failed ***"                             #bundle#
             pass
-        if focus is not None:
-            focus.focus()
         if timeout > 0:
             self.timer = after(self.top, timeout, self.mTimeout)
         try: self.top.mainloop()
@@ -101,8 +108,16 @@ oEEBEQ4qJKggAQA7"""))
     def destroy(self):
         after_cancel(self.timer)
         unbind_destroy(self.top)
-        self.top.wm_withdraw()
-        self.top.destroy()
+        try:
+            self.top.wm_withdraw()
+        except:
+            if traceback: traceback.print_exc()
+            pass
+        try:
+            self.top.destroy()
+        except:
+            if traceback: traceback.print_exc()
+            pass
         destruct(self.top)
         if self.parent:
             try:
@@ -112,14 +127,15 @@ oEEBEQ4qJKggAQA7"""))
                     self.parent.busyUpdate()
                 else:
                     self.parent.update()
-            except: pass
+            except:
+                if traceback: traceback.print_exc()
+                pass
+        self.top = None
+        self.parent = None
 
     def getDefaultFont(self):
         return getFont("default")
 
-    def mDefault(self, *event):
-        pass
-
     def wmDeleteWindow(self, *event):
         self.status = 1
         raise SystemExit
@@ -141,11 +157,11 @@ oEEBEQ4qJKggAQA7"""))
 class MfxDialog(_ToplevelDialog):
     def __init__(self, parent, title, **kw):
         kw = self.initKw(kw)
-        _ToplevelDialog.__init__(self, parent, title, kw.resizable)
+        _ToplevelDialog.__init__(self, parent, title, kw.resizable, kw.default)
         top_frame, bottom_frame = self.createFrames(kw)
         self.createBitmaps(top_frame, kw)
         #
-        self.num = kw.default
+        self.button = kw.default
         msg = Tkinter.Label(top_frame, text=kw.text, justify=kw.justify,
                             width=kw.width, font=kw.font)
         msg.pack(fill=Tkinter.BOTH, expand=1, padx=kw.padx, pady=kw.pady)
@@ -157,7 +173,7 @@ class MfxDialog(_ToplevelDialog):
         kw = KwStruct(kw,
             timeout=0, resizable=0,
             text="", justify="center",
-            strings=("Ok",), default=0,
+            strings=("OK",), default=0,
             width=0,
             font=self.getDefaultFont(),
             buttonfont=self.getDefaultFont(),
@@ -171,16 +187,16 @@ class MfxDialog(_ToplevelDialog):
         return kw
 
     def createFrames(self, kw):
-        frame = Tkinter.Frame(self.top)
-        frame.pack(side=Tkinter.TOP, fill=Tkinter.BOTH, expand=1)
+        bottom_frame = Tkinter.Frame(self.top)
+        bottom_frame.pack(side=Tkinter.BOTTOM, fill=Tkinter.BOTH, ipady=3)
         if kw.separatorwidth > 0:
             separator = Tkinter.Frame(self.top, relief="sunken",
                     height=kw.separatorwidth, width=kw.separatorwidth,
                     borderwidth=kw.separatorwidth / 2)
-            separator.pack(side=Tkinter.TOP, fill=Tkinter.X)
-        bot = Tkinter.Frame(self.top)
-        bot.pack(side=Tkinter.BOTTOM, fill=Tkinter.BOTH, ipady=3)
-        return frame, bot
+            separator.pack(side=Tkinter.BOTTOM, fill=Tkinter.X)
+        top_frame = Tkinter.Frame(self.top)
+        top_frame.pack(side=Tkinter.TOP, fill=Tkinter.BOTH, expand=1)
+        return top_frame, bottom_frame
 
     def createBitmaps(self, frame, kw):
         bm = ["error", "info", "questhead", "warning"]
@@ -195,30 +211,48 @@ class MfxDialog(_ToplevelDialog):
             b.pack(side=kw.image_side, padx=kw.image_padx, pady=kw.image_pady)
 
     def createButtons(self, frame, kw):
-        num = 0
+        button = column = -1
         padx, pady = kw.get("buttonpadx", 10), kw.get("buttonpady", 10)
         focus = None
+        max_len = 6
         for s in kw.strings:
-            b = Tkinter.Button(frame, text=s, font=kw.buttonfont,
-                               command=(lambda self=self, num=num: self.mDone(num)))
-            if num == kw.default:
-                b.config(default="active")
-                focus = b
+            if type(s) is types.TupleType:
+                s = s[0]
+            if s:
+                s = re.sub(r"[\s\.\,]", "", s)
+                max_len = max(max_len, len(s))
+        for s in kw.strings:
+            xbutton = button = button + 1
+            if type(s) is types.TupleType:
+                assert len(s) == 2
+                button = int(s[1])
+                s = s[0]
+            if s is None:
+                continue
+            if button < 0:
+                b = Tkinter.Button(frame, text=s, font=kw.buttonfont, state="disabled")
+                button = xbutton
             else:
-                b.config(default="normal")
-            b.grid_configure(column=num, row=0, sticky="ew", padx=padx, pady=pady)
-            b.grid_columnconfigure(num)
-            num = num + 1
-        if 0 <= kw.default < len(kw.strings):
-            bind(self.top, "<Return>", self.mDefault)
-            bind(self.top, "<KP_Enter>", self.mDefault)
+                b = Tkinter.Button(frame, text=s, font=kw.buttonfont, default="normal",
+                                   command=(lambda self=self, button=button: self.mDone(button)))
+                if button == kw.default:
+                    focus = b
+                    focus.config(default="active")
+            l = len(s)
+            if 1 and l < max_len:
+                l = l + (max_len - l) / 2
+                b.config(width=l)
+            column = column + 1
+            b.grid_configure(column=column, row=0, sticky="ew", padx=padx, pady=pady)
+            b.grid_columnconfigure(column)
+        if focus is not None:
+            l = (lambda event=None, self=self, button=kw.default: self.mDone(button))
+            bind(self.top, "<Return>", l)
+            bind(self.top, "<KP_Enter>", l)
         return focus
 
-    def mDefault(self, *event):
-        raise SystemExit
-
-    def mDone(self, num):
-        self.num = num
+    def mDone(self, button):
+        self.button = button
         raise SystemExit
 
 
@@ -234,11 +268,11 @@ class MfxExceptionDialog(MfxDialog):
             text = text + "\n"
         text = text + "\n"
         if isinstance(ex, EnvironmentError) and ex.filename is not None:
-            t = '[Errno %s] %s:\n%s' % (ex.errno, ex.strerror, repr(ex.filename))
+            t = "[Errno %s] %s:\n%s" % (ex.errno, ex.strerror, repr(ex.filename))
         else:
             t = str(ex)
         kw.text = text + t
-        apply(MfxDialog.__init__, (self, parent, title), kw.__dict__)
+        apply(MfxDialog.__init__, (self, parent, title), kw.getKw())
 
 
 # /***********************************************************************
@@ -249,11 +283,10 @@ class MfxSimpleSlider(MfxDialog):
     def __init__(self, parent, title,
                  label, value, from_, to, resolution, **kw):
         kw = self.initKw(kw)
-        _ToplevelDialog.__init__(self, parent, title, kw.resizable)
+        _ToplevelDialog.__init__(self, parent, title, kw.resizable, kw.default)
         top_frame, bottom_frame = self.createFrames(kw)
         self.createBitmaps(top_frame, kw)
         #
-        self.num = kw.default
         self.value = value
         self.var = Tkinter.DoubleVar()
         self.var.set(value)
@@ -267,17 +300,13 @@ class MfxSimpleSlider(MfxDialog):
 
     def initKw(self, kw):
         kw = KwStruct(kw,
-            strings=("Ok", "Cancel"), default=0,
+            strings=("OK", "Cancel"), default=0,
             separatorwidth = 0,
         )
         return MfxDialog.initKw(self, kw)
 
-    def mDefault(self, *event):
-        self.value = self.var.get()
-        raise SystemExit
-
-    def mDone(self, num):
-        self.num = num
+    def mDone(self, button):
+        self.button = button
         self.value = self.var.get()
         raise SystemExit
 
@@ -289,11 +318,10 @@ class MfxSimpleSlider(MfxDialog):
 class MfxSimpleEntry(MfxSimpleSlider):
     def __init__(self, parent, title, label, value, **kw):
         kw = self.initKw(kw)
-        _ToplevelDialog.__init__(self, parent, title, kw.resizable)
+        _ToplevelDialog.__init__(self, parent, title, kw.resizable, kw.default)
         top_frame, bottom_frame = self.createFrames(kw)
         self.createBitmaps(top_frame, kw)
         #
-        self.num = kw.default
         self.value = value
         if label:
             label = Tkinter.Label(top_frame, text=label, takefocus=0)
@@ -325,9 +353,10 @@ class MfxTooltip:
         self.bindings.append(self.widget.bind("<Leave>", self._leave))
         self.bindings.append(self.widget.bind("<ButtonPress>", self._leave))
         # user overrideable settings
-        self.time = 1500                    # milliseconds
+        self.time = 1000                    # milliseconds
         self.relief = Tkinter.SOLID
         self.justify = Tkinter.LEFT
+        self.fg = "#000000"
         self.bg = "#ffffe0"
         self.xoffset = 20
         self.yoffset = 1
@@ -378,66 +407,165 @@ class MfxTooltip:
         self.tooltip.wm_protocol("WM_DELETE_WINDOW", self.destroy)
         self.label = Tkinter.Label(self.tooltip, text=self.text,
                          relief=self.relief, justify=self.justify,
-                         bg=self.bg, bd=1, takefocus=0)
+                         fg=self.fg, bg=self.bg, bd=1, takefocus=0)
         self.label.pack(ipadx=1, ipady=1)
-        if os.name != "nt":
-            self.tooltip.wm_geometry("%+d%+d" % (x, y))
+        self.tooltip.wm_geometry("%+d%+d" % (x, y))
         self.tooltip.wm_deiconify()
-        if os.name == "nt":
-            self.tooltip.wm_geometry("%+d%+d" % (x, y))
         ##self.tooltip.tkraise()
 
 
 # /***********************************************************************
-# // A canvas widget with scroll bars and some useful bindings.
-# // Adapted from IDLE.
+# // A canvas widget with scrollbars and some useful bindings.
 # ************************************************************************/
 
 class MfxScrolledCanvas:
-    def __init__(self, parent, vbar=1, hbar=0, **kw):
-        kwdefault(kw, bg=parent.cget("bg"), highlightthickness=0,
+    def __init__(self, parent, hbar=2, vbar=2, **kw):
+        bg = kw.get("bg", parent.cget("bg"))
+        kwdefault(kw, bg=bg, highlightthickness=0,
                   xscrollincrement=16, yscrollincrement=16)
-        bg = kw["bg"]
         self.parent = parent
-        self.frame = Tkinter.Frame(parent, bg=bg)
+        self.createFrame(kw)
+        self.canvas = None
+        self.hbar = None
+        self.hbar_mode = hbar
+        self.vbar = None
+        self.vbar_mode = vbar
+        self.hbar_show = 0
+        self.vbar_show = 0
+        self.resize_pending = 0
+        self.timer = None
+        self.createCanvas(kw)
         self.frame.grid_rowconfigure(0, weight=1)
         self.frame.grid_columnconfigure(0, weight=1)
-        self.canvas = None
-        self.vbar = None
-        self.hbar = None
-        self.createCanvas(kw)
-        if vbar:
-            self.createVbar(bg)
         if hbar:
+            if hbar == 3:
+                w = 21
+                if win32api:
+                    w = win32api.GetSystemMetrics(3)        # SM_CYHSCROLL
+                self.frame.grid_rowconfigure(1, minsize=w)
             self.createHbar(bg)
-        self.canvas.focus_set()
+            if not vbar:
+                bind(self.hbar, "<Map>", self._mapBar)
+            self.bindHbar()
+        if vbar:
+            if vbar == 3:
+                w = 21
+                if win32api:
+                    w = win32api.GetSystemMetrics(2)        # SM_CXVSCROLL
+                self.frame.grid_columnconfigure(1, minsize=w)
+            self.createVbar(bg)
+            bind(self.vbar, "<Map>", self._mapBar)
+            self.bindVbar()
+        ###self.canvas.focus_set()
 
-    def unbind(self):
+    #
+    #
+    #
+
+    def destroy(self):
+        after_cancel(self.timer)
+        self.timer = None
+        self.unbind_all()
+        self.canvas.destroy()
+        self.frame.destroy()
+
+    def pack(self, **kw):
+        apply(self.frame.pack, (), kw)
+
+    #
+    #
+    #
+
+    def unbind_all(self):
+        unbind_destroy(self.hbar)
+        unbind_destroy(self.vbar)
         unbind_destroy(self.canvas)
         unbind_destroy(self.frame)
 
+    def createFrame(self, kw):
+        width = kw.get("width")
+        height = kw.get("height")
+        self.frame = Tkinter.Frame(self.parent, width=width, height=height, bg=None)
     def createCanvas(self, kw):
         self.canvas = apply(Tkinter.Canvas, (self.frame,), kw)
-        self.canvas.grid(row=0, column=0, sticky="nsew")
-    def createVbar(self, bg):
-        self.vbar = Tkinter.Scrollbar(self.frame, name="vbar", bg=bg, takefocus=0)
-        self.vbar.grid(row=0, column=1, sticky="nse")
-        self.canvas["yscrollcommand"] = self.vbar.set
-        self.vbar["command"] = self.canvas.yview
-        bind(self.canvas, "<KeyPress-Prior>", self.page_up)
-        bind(self.canvas, "<KeyPress-Next>", self.page_down)
-        bind(self.canvas, "<KeyPress-Up>", self.unit_up)
-        bind(self.canvas, "<KeyPress-Down>", self.unit_down)
-        bind(self.canvas, "<KeyPress-Begin>", self.scroll_top)
-        bind(self.canvas, "<KeyPress-Home>", self.scroll_top)
-        bind(self.canvas, "<KeyPress-End>", self.scroll_bottom)
+        self.canvas.grid(row=0, column=0, sticky="news")
     def createHbar(self, bg):
         self.hbar = Tkinter.Scrollbar(self.frame, name="hbar", bg=bg, takefocus=0, orient="horizontal")
-        self.hbar.grid(row=1, column=0, sticky="ews")
-        self.canvas["xscrollcommand"] = self.hbar.set
+        self.canvas["xscrollcommand"] = self._setHbar
         self.hbar["command"] = self.canvas.xview
-        bind(self.canvas, "<KeyPress-Left>", self.unit_left)
-        bind(self.canvas, "<KeyPress-Right>", self.unit_right)
+    def createVbar(self, bg):
+        self.vbar = Tkinter.Scrollbar(self.frame, name="vbar", bg=bg, takefocus=0)
+        self.canvas["yscrollcommand"] = self._setVbar
+        self.vbar["command"] = self.canvas.yview
+    def bindHbar(self, w=None):
+        if w is None:
+            w = self.canvas
+        bind(w, "<KeyPress-Left>", self.unit_left)
+        bind(w, "<KeyPress-Right>", self.unit_right)
+    def bindVbar(self, w=None):
+        if w is None:
+            w = self.canvas
+        bind(w, "<KeyPress-Prior>", self.page_up)
+        bind(w, "<KeyPress-Next>", self.page_down)
+        bind(w, "<KeyPress-Up>", self.unit_up)
+        bind(w, "<KeyPress-Down>", self.unit_down)
+        bind(w, "<KeyPress-Begin>", self.scroll_top)
+        bind(w, "<KeyPress-Home>", self.scroll_top)
+        bind(w, "<KeyPress-End>", self.scroll_bottom)
+
+    def _mapBar(self, event):
+        # see: autoscroll.tcl, http://mini.net/cgi-bin/wikit/950.html
+        top = event.widget.winfo_toplevel()
+        g = top.wm_geometry()
+        if self.resize_pending:
+            self.resize_pending = 0
+            self.canvas.update_idletasks()
+        top.wm_geometry(g)
+
+    def _setHbar(self, *args):
+        apply(self.hbar.set, args)
+        self.showHbar()
+    def _setVbar(self, *args):
+        apply(self.vbar.set, args)
+        self.showVbar()
+
+    def showHbar(self, show=-1):
+        if not self.hbar:
+            return 0
+        if show < 0:
+            show = self.hbar_mode
+        if show > 1:
+            if not self.canvas.winfo_ismapped():
+                return 0
+            view = self.canvas.xview()
+            show = abs(view[0]) > 0.0001 or abs(view[1] - 1.0) > 0.0001
+        if show == self.hbar_show:
+            return 0
+        if show:
+            self.hbar.grid(row=1, column=0, sticky="we")
+        else:
+            self.hbar.grid_forget()
+        self.hbar_show = show
+        return 1
+
+    def showVbar(self, show=-1):
+        if not self.vbar:
+            return 0
+        if show < 0:
+            show = self.vbar_mode
+        if show > 1:
+            if not self.canvas.winfo_ismapped():
+                return 0
+            view = self.canvas.yview()
+            show = abs(view[0]) > 0.0001 or abs(view[1] - 1.0) > 0.0001
+        if show == self.vbar_show:
+            return 0
+        if show:
+            self.vbar.grid(row=0, column=1, sticky="ns")
+        else:
+            self.vbar.grid_forget()
+        self.vbar_show = show
+        return 1
 
     def page_up(self, *event):
         self.canvas.yview_scroll(-1, "page")
@@ -478,14 +606,14 @@ class MfxScrolledCanvas:
 #%ifndef BUNDLE
 
 def tkwidget_main(args):
-    from tkutil import withdraw                                     #bundle#
+    from tkutil import wm_withdraw
     tk = Tkinter.Tk()
-    withdraw(tk)
+    wm_withdraw(tk)
     tk.update()
     if 1:
         d = MfxDialog(tk, "title", text="This is an info dialog",
-                      bitmap="info", default=0, strings=("Ok",))
-        print d.status, d.num
+                      bitmap="info", default=0, strings=("OK",))
+        print d.status, d.button
     if 1:
         d = MfxSimpleSlider(tk, "title", "label", 1.0, 0.2, 9.9, 0.1)
         print d.status, d.value
